#    Copyright (C) 2020 David J. Wooten
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np

from ..single import Hill
from .. import utils
from .nonparametric_base import DoseDependentModel

# Used for delta variant of Loewe
from scipy.optimize import minimize_scalar

class Loewe(DoseDependentModel):
    """Loewe Additivity Synergy
    
    Loewe's model of drug combination additivity expects a linear tradeoff, such that withholding X parts of drug 1 can be compensated for by adding Y parts of drug 2. In Loewe's model, X and Y are constant (e.g., withholding 5X parts of drug 1 will be compensated for with 5Y parts of drug 2.)

    We support two methods to calculate Loewe synergy. For each method, synergy and antagonism are defined and interpreted differently (see "synergy" below).
    
    The vairant may be set as 
    
        model = Loewe(..., variant="CI") (default)
        or
        model = Loewe(..., variant="delta")

    Variant "CI" (default) calculates synergy using an equation equivalent to combination index, but unlike CI, Loewe allows arbitrary single-drug models.

    Variant "delta" calculates an expected dose response surface, and defines synergy as the difference between the measured values and the expected values. This variant REQUIRES single-drugs be fitted using a Hill model.

    synergy : array_like, float
        if variant=="CI":
            [0,1)=synergism, (1,inf)=antagonism
        if variant=="delta":
            (0,inf)=synergism, (-inf,0)=antagonism
    """

    def __init__(self, h1_bounds=(0,np.inf), h2_bounds=(0,np.inf),  \
            C1_bounds=(0,np.inf), C2_bounds=(0,np.inf),             \
            E0_bounds=(-np.inf,np.inf), E1_bounds=(-np.inf,np.inf), \
            E2_bounds=(-np.inf,np.inf), variant="CI"):
        
        super().__init__(h1_bounds=h1_bounds, h2_bounds=h2_bounds, C1_bounds=C1_bounds, C2_bounds=C2_bounds, E0_bounds=E0_bounds, E1_bounds=E1_bounds, E2_bounds=E2_bounds)
        
        self.variant = "CI"
        if isinstance(variant, str):
            self.variant = variant.lower()

    def fit(self, d1, d2, E, drug1_model=None, drug2_model=None, **kwargs):

        d1 = np.asarray(d1)
        d2 = np.asarray(d2)
        E = np.asarray(E)
        super().fit(d1,d2,E, drug1_model=drug1_model, drug2_model=drug2_model, **kwargs)

        drug1_model = self.drug1_model
        drug2_model = self.drug2_model


        if self.variant.startswith("delta"):
            self.reference = self._E_reference(d1, d2, drug1_model, drug2_model)
            self.synergy = self._get_synergy_delta(d1, d2, E, drug1_model, drug2_model)
        else:
            self.synergy = self._get_synergy_CI(d1, d2, E, drug1_model, drug2_model)

        return self.synergy
    
    def _get_synergy_delta(self, d1, d2, E, drug1_model, drug2_model):
        
        # Save reference and synergy
        if self.reference is None:
            self.reference = self._E_reference(d1, d2, drug1_model, drug2_model)
        synergy = self.reference - E

        if hasattr(synergy,"__iter__"): synergy[(d1==0) | (d2==0)] = 0
        elif d1==0 or d2==0: synergy=0
        return synergy

    def _get_synergy_CI(self, d1, d2, E, drug1_model, drug2_model):
        with np.errstate(divide='ignore', invalid='ignore'):
            d1_alone = drug1_model.E_inv(E)
            d2_alone = drug2_model.E_inv(E)
            synergy = d1/d1_alone + d2/d2_alone
        
        if hasattr(synergy,"__iter__"): synergy[(d1==0) | (d2==0)] = 1
        elif d1==0 or d2==0: synergy=1
        return synergy

    def _get_single_drug_classes(self):
        # The delta model ONLY works when single drugs are fit with Hill equation
        #if self.variant.startswith("delta"):
        #    return Hill, Hill

        # Otherwise, default to Hill, but allow anything
        return Hill, None

    def _fit_Loewe_reference(self, X_r, X_c, drug1_model, drug2_model):
        """Calculates a reference (null) model for Loewe for drug1 and drug2 at concentrations X_r and X_c

        drug1_model and drug2_model MUST be some form of Hill equation
        
        Contributed by Mark Russo at Bristol Myers Squibb

        Returns scipy.optimize.minimize_scalar object
        """
        # Unpack single drug parameters
        #pa_r = drug1_model.get_parameters()
        #pa_c = drug2_model.get_parameters()
        #[Emin_r, Emax_r, h_r, m_r] = pa_r
        #[Emin_c, Emax_c, h_c, m_c] = pa_c

        # Compute the bounds within which Y_Loewe is valid.
        # Any value outside these bounds causes algorithm to take a root of a negative.
        #bounds_r = [Emin_r, Emax_r]; bounds_r.sort()
        #bounds_c = [Emin_c, Emax_c]; bounds_c.sort()
        #bounds   = [max(bounds_r[0], bounds_c[0]), min(bounds_r[1], bounds_c[1])]


        
        bounds1 = sorted([drug1_model.E0, drug1_model.Emax])
        bounds2 = sorted([drug2_model.E0, drug2_model.Emax])

        bounds   = [max(bounds1[0], bounds2[0]), min(bounds1[1], bounds2[1])]
        


        # Quadratic function with a minimum at CI=1
        #f  = lambda Y: ((X_r/(m_r*(((Y-Emin_r)/(Emax_r-Y))**(1/h_r))) + X_c/(m_c*(((Y-Emin_c)/(Emax_c-Y))**(1/h_c)))) - 1.0)**2

        f = lambda Y : ((X_r / drug1_model.E_inv(Y)) + (X_c / drug2_model.E_inv(Y)) -1.0)**2

        # Perform constrained minimization to find Y as close as possible to CI == 1 and return result.
        return minimize_scalar(f, method='bounded', bounds=bounds)

    def _E_reference(self, d1, d2, drug1_model, drug2_model):
        """Calculates a reference (null) model for Loewe for drug1 and drug2.
        
        Contributed by Mark Russo at Bristol Myers Squibb
        Modified by David Wooten
        """

        # Compute the reference model
        with np.errstate(divide='ignore', invalid='ignore'):
            ref  = 0*d1

            #pa1 = drug1_model.get_parameters()
            #pa2 = drug2_model.get_parameters()

            Emax_1 = drug1_model.Emax
            Emax_2 = drug2_model.Emax

            #weakest_E = max(pa1[1], pa2[1])
            weakest_E = max(Emax_1, Emax_2)

            stronger_drug = drug1_model
            #if pa1[1] > pa2[1]: # pa[1] is Emax
            if Emax_1 > Emax_2: # pa[1] is Emax
                stronger_drug = drug2_model

            # Loewe becomes undefined for effects past the weaker drug's Emax
            # We implement several variants to handle this case:
            #  1) variant="delta" - this will set E_reference to weakest_E
            #  2) variant="delta_HSA" - this will set E_r to min(E1, E2)
            #  3) variant="delta_nan" - this will set E_r to nan
            #  4) variant="synergyfinder" - sets E_r to stronger.E(d1+d2)
            if self.variant=="delta": option=1
            elif self.variant=="delta_hsa": option=2
            elif self.variant=="delta_nan": option=3
            elif self.variant=="delta_synergyfinder": option=4
            else: option=1

            for i in range(len(ref)):
                X1, X2 = d1[i], d2[i]
                E1, E2 = drug1_model.E(X1), drug2_model.E(X2)
                
                # No drug
                if X1 == 0.0 and X2 == 0.0:
                    ref[i] = 0.5 * (E2 + E1)

                # Single drugs
                elif X2 == 0:
                    ref[i] = E1
                elif X1 == 0:
                    ref[i] = E2

                # If the combo E is stronger than the weaker drug is capable of, don't bother trying to fit
                elif E1 < weakest_E or E2 < weakest_E:
                    if option==1:
                        ref[i] = weakest_E
                    elif option==2:
                        ref[i] = min(E1,E2)
                    elif option==3:
                        ref[i] = np.nan
                    elif option==4:
                        ref[i] = stronger_drug.E(X1+X2)
                    else:
                        ref[i] = np.nan
                else:
                    # Numerically solve the value for Loewe
                    res = self._fit_Loewe_reference(X1, X2, drug1_model, drug2_model)
                    if res.success:
                        ref[i] = res.x
                    else:
                        ref[i] = min(E2, E1)
        return ref

    def plot_heatmap(self, cmap="PRGn", neglog=None, center_on_zero=True, **kwargs):
        if neglog is None:
            if self.variant=="delta":
                neglog=False
            else:
                neglog=True
        super().plot_heatmap(cmap=cmap, neglog=neglog, center_on_zero=center_on_zero, **kwargs)

    def plot_surface_plotly(self, cmap="PRGn", neglog=None, center_on_zero=True, **kwargs):
        if neglog is None:
            if self.variant=="delta":
                neglog=False
            else:
                neglog=True
        super().plot_surface_plotly(cmap=cmap, neglog=neglog, center_on_zero=center_on_zero, **kwargs)


#    Copyright (C) 2020 David J. Wooten
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np

from .jacobians.musyc_jacobian import jacobian
from .parametric_base import ParametricModel

from .. import utils
from ..single import Hill

class MuSyC(ParametricModel):
    """Multidimensional Synergy of Combinations (MuSyC) is a drug synergy framework based on the law of mass action (doi: 10.1016/j.cels.2019.01.003, doi: 10.1101/683433). In MuSyC, synergy is parametrically defined as shifts in potency, efficacy, or cooperativity.

    alpha21 : float
        Synergistic potency ([0,1) = antagonism, (1,inf) = synergism).        At large concentrations of drug 2, the "effective dose" of drug 1 = alpha21*d1.
    
    alpha12 : float
        Synergistic potency ([0,1) = antagonism, (1,inf) = synergism).         At large concentrations of drug 1, the "effective dose" of drug 2 = alpha12*d2.

    beta : float
        Synergistic efficacy ((-inf,0) = antagonism, (0,inf) = synergism). At large concentrations of both drugs, the combination achieves an effect beta-% stronger (or weaker) than the stronger single-drug.

    gamma21 : float
        Synergistic cooperativity ([0,1) = antagonism, (1,inf) = synergism). At large concentrations of drug 2, the Hill slope of drug 1 = gamma21*h1

    gamma12 : float
        Synergistic cooperativity ([0,1) = antagonism, (1,inf) = synergism). At large concentrations of drug 1, the Hill slope of drug 2 = gamma12*h2

    """
    def __init__(self, h1_bounds=(0,np.inf), h2_bounds=(0,np.inf),  \
            C1_bounds=(0,np.inf), C2_bounds=(0,np.inf),             \
            E0_bounds=(-np.inf,np.inf), E1_bounds=(-np.inf,np.inf), \
            E2_bounds=(-np.inf,np.inf), E3_bounds=(-np.inf,np.inf), \
            alpha12_bounds=(0,np.inf), alpha21_bounds=(0,np.inf),   \
            gamma12_bounds=(0,np.inf), gamma21_bounds=(0,np.inf),   \
            r1r=1., r2r=1., E0=None, E1=None, E2=None, E3=None,     \
            h1=None, h2=None, C1=None, C2=None, alpha12=None,       \
            alpha21=None, gamma12=None, gamma21=None, variant="full"):
        super().__init__()
        self.C1_bounds = C1_bounds
        self.C2_bounds = C2_bounds
        self.h1_bounds = h1_bounds
        self.h2_bounds = h2_bounds
        self.E0_bounds = E0_bounds
        self.E1_bounds = E1_bounds
        self.E2_bounds = E2_bounds
        self.E3_bounds = E3_bounds
        self.alpha12_bounds = alpha12_bounds
        self.alpha21_bounds = alpha21_bounds
        self.gamma12_bounds = gamma12_bounds
        self.gamma21_bounds = gamma21_bounds

        self.variant = variant

        self.r1r = r1r
        self.r2r = r2r
        self.E0 = E0
        self.E1 = E1
        self.E2 = E2
        self.E3 = E3
        self.h1 = h1
        self.h2 = h2
        self.C1 = C1
        self.C2 = C2
        self.alpha12 = alpha12
        self.alpha21 = alpha21
        self.gamma12 = gamma12
        self.gamma21 = gamma21
        if not None in [E1, E2, E3]:
            self.beta = (min(E1,E2)-E3) / (E0 - min(E1,E2))
        else:
            self.beta = None

        with np.errstate(divide='ignore'):
            self.logh1_bounds = (np.log(h1_bounds[0]), np.log(h1_bounds[1]))
            self.logC1_bounds = (np.log(C1_bounds[0]), np.log(C1_bounds[1]))
            self.logh2_bounds = (np.log(h2_bounds[0]), np.log(h2_bounds[1]))
            self.logC2_bounds = (np.log(C2_bounds[0]), np.log(C2_bounds[1]))
            
            self.logalpha12_bounds = (np.log(alpha12_bounds[0]), np.log(alpha12_bounds[1]))
            self.logalpha21_bounds = (np.log(alpha21_bounds[0]), np.log(alpha21_bounds[1]))

            self.loggamma12_bounds = (np.log(gamma12_bounds[0]), np.log(gamma12_bounds[1]))
            self.loggamma21_bounds = (np.log(gamma21_bounds[0]), np.log(gamma21_bounds[1]))


        if variant == "full":
            self.fit_function = lambda d, E0, E1, E2, E3, logh1, logh2, logC1, logC2, logalpha12, logalpha21, loggamma12, loggamma21: self._model(d[0], d[1], E0, E1, E2, E3, np.exp(logh1), np.exp(logh2), np.exp(logC1), np.exp(logC2), self.r1r, self.r2r, np.exp(logalpha12), np.exp(logalpha21), np.exp(loggamma12), np.exp(loggamma21))

            self.jacobian_function = lambda d, E0, E1, E2, E3, logh1, logh2, logC1, logC2, logalpha12, logalpha21, loggamma12, loggamma21: jacobian(d[0], d[1], E0, E1, E2, E3, logh1, logh2, logC1, logC2, self.r1r, self.r2r, logalpha12, logalpha21, loggamma12, loggamma21)

            self.bounds = tuple(zip(self.E0_bounds, self.E1_bounds, self.E2_bounds, self.E3_bounds, self.logh1_bounds, self.logh2_bounds, self.logC1_bounds, self.logC2_bounds, self.logalpha12_bounds, self.logalpha21_bounds, self.loggamma12_bounds, self.loggamma21_bounds))
        
        elif variant == "no_gamma":
            self.fit_function = lambda d, E0, E1, E2, E3, logh1, logh2, logC1, logC2, logalpha12, logalpha21: self._model(d[0], d[1], E0, E1, E2, E3, np.exp(logh1), np.exp(logh2), np.exp(logC1), np.exp(logC2), self.r1r, self.r2r, np.exp(logalpha12), np.exp(logalpha21), 1, 1)

            self.jacobian_function = lambda d, E0, E1, E2, E3, logh1, logh2, logC1, logC2, logalpha12, logalpha21: jacobian(d[0], d[1], E0, E1, E2, E3, logh1, logh2, logC1, logC2, self.r1r, self.r2r, logalpha12, logalpha21, 0, 0)[:,:-2]

            self.bounds = tuple(zip(self.E0_bounds, self.E1_bounds, self.E2_bounds, self.E3_bounds, self.logh1_bounds, self.logh2_bounds, self.logC1_bounds, self.logC2_bounds, self.logalpha12_bounds, self.logalpha21_bounds))


    def _get_initial_guess(self, d1, d2, E, drug1_model, drug2_model, p0=None):
        

        # If there is no intial guess, use single-drug models to come up with intitial guess
        if p0 is None:
            # Sanitize single-drug models
            default_class, expected_superclass = self._get_single_drug_classes()

            drug1_model = utils.sanitize_single_drug_model(drug1_model, default_class, expected_superclass=expected_superclass, E0_bounds=self.E0_bounds, Emax_bounds=self.E1_bounds, h_bounds=self.h1_bounds, C_bounds=self.C1_bounds)

            drug2_model = utils.sanitize_single_drug_model(drug2_model, default_class, expected_superclass=expected_superclass, E0_bounds=self.E0_bounds, Emax_bounds=self.E2_bounds, h_bounds=self.h2_bounds, C_bounds=self.C2_bounds)

            # Fit the single drug models if they were not pre-fit by the user
            if not drug1_model.is_fit():
                mask = np.where(d2==min(d2))
                drug1_model.fit(d1[mask], E[mask])
            if not drug2_model.is_fit():
                mask = np.where(d1==min(d1))
                drug2_model.fit(d2[mask], E[mask])

            # Get initial guesses of E0, E1, E2, h1, h2, C1, and C2 from single-drug fits
            E0_1, E1, h1, C1 = drug1_model.get_parameters()
            E0_2, E2, h2, C2 = drug2_model.get_parameters()
            
            #TODO: E orientation
            # Get initial guess of E3 at E(d1_max, d2_max), if that point exists
            E3 = E[(d1==max(d1)) & (d2==max(d2))]
            if len(E3)>0: E3 = np.mean(E3)

            # Otherwise guess E3 is the minimum E observed
            else: E3 = np.min(E)
            
            p0 = [(E0_1+E0_2)/2., E1, E2, E3, h1, h2, C1, C2, 1, 1, 1, 1]
        
            if self.variant == "no_gamma":
                p0 = p0[:-2]

        p0 = list(self._transform_params_to_fit(p0))
        utils.sanitize_initial_guess(p0, self.bounds)
        return p0

    def _transform_params_from_fit(self, params):
        
        if self.variant == "no_gamma":
            E0, E1, E2, E3, logh1, logh2, logC1, logC2, logalpha12, logalpha21 = params
        else:
            E0, E1, E2, E3, logh1, logh2, logC1, logC2, logalpha12, logalpha21, loggamma12, loggamma21 = params
            gamma12 = np.exp(loggamma12)
            gamma21 = np.exp(loggamma21)
        
        h1 = np.exp(logh1)
        h2 = np.exp(logh2)
        C1 = np.exp(logC1)
        C2 = np.exp(logC2)
        alpha12 = np.exp(logalpha12)
        alpha21 = np.exp(logalpha21)
        
        if self.variant == "no_gamma":
            return E0, E1, E2, E3, h1, h2, C1, C2, alpha12, alpha21
        
        return E0, E1, E2, E3, h1, h2, C1, C2, alpha12, alpha21, gamma12, gamma21

    def _transform_params_to_fit(self, params):
        
        if self.variant == "no_gamma":
            E0, E1, E2, E3, h1, h2, C1, C2, alpha12, alpha21 = params
        else:
            E0, E1, E2, E3, h1, h2, C1, C2, alpha12, alpha21, gamma12, gamma21 = params
            loggamma12 = np.log(gamma12)
            loggamma21 = np.log(gamma21)

        logh1 = np.log(h1)
        logh2 = np.log(h2)
        logC1 = np.log(C1)
        logC2 = np.log(C2)
        logalpha12 = np.log(alpha12)
        logalpha21 = np.log(alpha21)

        if self.variant == "no_gamma":
            return E0, E1, E2, E3, logh1, logh2, logC1, logC2, logalpha12, logalpha21

        return E0, E1, E2, E3, logh1, logh2, logC1, logC2, logalpha12, logalpha21, loggamma12, loggamma21

    def E(self, d1, d2):
        if not self._is_parameterized():
            return None

        if self.variant == "no_gamma":
            return self._model(d1, d2, self.E0, self.E1, self.E2, self.E3, self.h1, self.h2, self.C1, self.C2, self.r1r, self.r2r, self.alpha12, self.alpha21, 1, 1)
        
        else:
            return self._model(d1, d2, self.E0, self.E1, self.E2, self.E3, self.h1, self.h2, self.C1, self.C2, self.r1r, self.r2r, self.alpha12, self.alpha21, self.gamma12, self.gamma21)

    def _get_parameters(self):
        if self.variant == "no_gamma":
            return self.E0, self.E1, self.E2, self.E3, self.h1, self.h2, self.C1, self.C2, self.alpha12, self.alpha21
        else:
            return self.E0, self.E1, self.E2, self.E3, self.h1, self.h2, self.C1, self.C2, self.alpha12, self.alpha21, self.gamma12, self.gamma21
    
    def _set_parameters(self, popt):
        if self.variant == "no_gamma":
            self.E0, self.E1, self.E2, self.E3, self.h1, self.h2, self.C1, self.C2, self.alpha12, self.alpha21 = popt
        else:
            self.E0, self.E1, self.E2, self.E3, self.h1, self.h2, self.C1, self.C2, self.alpha12, self.alpha21, self.gamma12, self.gamma21 = popt

    def _get_single_drug_classes(self):
        return Hill, Hill

    def _C_to_r1(self, C, h, r1r):
        return r1r/np.power(C,h)

    @DeprecationWarning
    def _C_to_r1r(self, C, h, r1):
        return r1*C**h

    @DeprecationWarning
    def _r_to_C(self, h, r1r):
        return (r1r/r1)**(1./h)

    def _reference_E(self, d1, d2):
        if not self._is_parameterized():
            return None
        return self._model(d1, d2, self.E0, self.E1, self.E2, min(self.E1,self.E2), self.h1, self.h2, self.C1, self.C2, self.r1r, self.r2r, 1, 1, 1, 1)

    def _model(self, d1, d2, E0, E1, E2, E3, h1, h2, C1, C2, r1r, r2r, alpha12, alpha21, gamma12, gamma21):

        d1h1 = np.power(d1,h1)
        d2h2 = np.power(d2,h2)

        C1h1 = np.power(C1,h1)
        C2h2 = np.power(C2,h2)

        r1 = r1r/C1h1
        r2 = r2r/C2h2

        alpha21d1gamma21h1 = np.power(alpha21*d1, gamma21*h1)
        alpha12d2gamma12h2 = np.power(alpha12*d2, gamma12*h2)

        C12h1 = np.power(C1,2*h1)
        C22h2 = np.power(C2,2*h2)

        # ********** U ********

        U=(r1*r2*np.power((r1*C1h1),gamma21)*C1h1*C2h2+r1*r2*np.power((r2*C2h2),gamma12)*C1h1*C2h2+np.power(r1,(gamma21+1))*alpha21d1gamma21h1*np.power((r2*C2h2),gamma12)*C1h1+np.power(r2,(gamma12+1))*alpha12d2gamma12h2*np.power((r1*C1h1),gamma21)*C2h2)/(d1h1*r1*r2*np.power((r1*C1h1),gamma21)*C2h2+d1h1*r1*r2*np.power((r2*C2h2),gamma12)*C2h2+d1h1*r1*np.power(r2,(gamma12+1))*alpha12d2gamma12h2*C2h2+d1h1*r1*np.power(r2,gamma12)*alpha12d2gamma12h2*np.power((r1*C1h1),gamma21)+d1h1*np.power(r1,(gamma21+1))*np.power(r2,gamma12)*alpha21d1gamma21h1*alpha12d2gamma12h2+d1h1*np.power(r1,(gamma21+1))*alpha21d1gamma21h1*np.power((r2*C2h2),gamma12)+d2h2*r1*r2*np.power((r1*C1h1),gamma21)*C1h1+d2h2*r1*r2*np.power((r2*C2h2),gamma12)*C1h1+d2h2*np.power(r1,(gamma21+1))*r2*alpha21d1gamma21h1*C1h1+d2h2*np.power(r1,gamma21)*r2*alpha21d1gamma21h1*np.power((r2*C2h2),gamma12)+d2h2*np.power(r1,gamma21)*np.power(r2,(gamma12+1))*alpha21d1gamma21h1*alpha12d2gamma12h2+d2h2*np.power(r2,(gamma12+1))*alpha12d2gamma12h2*np.power((r1*C1h1),gamma21)+r1*r2*np.power((r1*C1h1),gamma21)*C1h1*C2h2+r1*r2*np.power((r2*C2h2),gamma12)*C1h1*C2h2+np.power(r1,(gamma21+1))*alpha21d1gamma21h1*np.power((r2*C2h2),gamma12)*C1h1+np.power(r2,(gamma12+1))*alpha12d2gamma12h2*np.power((r1*C1h1),gamma21)*C2h2)

        #**********E1********

        A1=(d1h1*r1*r2*np.power((r1*C1h1),gamma21)*C2h2+d1h1*r1*r2*np.power((r2*C2h2),gamma12)*C2h2+d1h1*np.power(r1,(gamma21+1))*alpha21d1gamma21h1*np.power((r2*C2h2),gamma12)+d2h2*np.power(r1,gamma21)*r2*alpha21d1gamma21h1*np.power((r2*C2h2),gamma12))/(d1h1*r1*r2*np.power((r1*C1h1),gamma21)*C2h2+d1h1*r1*r2*np.power((r2*C2h2),gamma12)*C2h2+d1h1*r1*np.power(r2,(gamma12+1))*alpha12d2gamma12h2*C2h2+d1h1*r1*np.power(r2,gamma12)*alpha12d2gamma12h2*np.power((r1*C1h1),gamma21)+d1h1*np.power(r1,(gamma21+1))*np.power(r2,gamma12)*alpha21d1gamma21h1*alpha12d2gamma12h2+d1h1*np.power(r1,(gamma21+1))*alpha21d1gamma21h1*np.power((r2*C2h2),gamma12)+d2h2*r1*r2*np.power((r1*C1h1),gamma21)*C1h1+d2h2*r1*r2*np.power((r2*C2h2),gamma12)*C1h1+d2h2*np.power(r1,(gamma21+1))*r2*alpha21d1gamma21h1*C1h1+d2h2*np.power(r1,gamma21)*r2*alpha21d1gamma21h1*np.power((r2*C2h2),gamma12)+d2h2*np.power(r1,gamma21)*np.power(r2,(gamma12+1))*alpha21d1gamma21h1*alpha12d2gamma12h2+d2h2*np.power(r2,(gamma12+1))*alpha12d2gamma12h2*np.power((r1*C1h1),gamma21)+r1*r2*np.power((r1*C1h1),gamma21)*C1h1*C2h2+r1*r2*np.power((r2*C2h2),gamma12)*C1h1*C2h2+np.power(r1,(gamma21+1))*alpha21d1gamma21h1*np.power((r2*C2h2),gamma12)*C1h1+np.power(r2,(gamma12+1))*alpha12d2gamma12h2*np.power((r1*C1h1),gamma21)*C2h2)

        #**********E2********

        A2=(d1h1*r1*np.power(r2,gamma12)*alpha12d2gamma12h2*np.power((r1*C1h1),gamma21)+d2h2*r1*r2*np.power((r1*C1h1),gamma21)*C1h1+d2h2*r1*r2*np.power((r2*C2h2),gamma12)*C1h1+d2h2*np.power(r2,(gamma12+1))*alpha12d2gamma12h2*np.power((r1*C1h1),gamma21))/(d1h1*r1*r2*np.power((r1*C1h1),gamma21)*C2h2+d1h1*r1*r2*np.power((r2*C2h2),gamma12)*C2h2+d1h1*r1*np.power(r2,(gamma12+1))*alpha12d2gamma12h2*C2h2+d1h1*r1*np.power(r2,gamma12)*alpha12d2gamma12h2*np.power((r1*C1h1),gamma21)+d1h1*np.power(r1,(gamma21+1))*np.power(r2,gamma12)*alpha21d1gamma21h1*alpha12d2gamma12h2+d1h1*np.power(r1,(gamma21+1))*alpha21d1gamma21h1*np.power((r2*C2h2),gamma12)+d2h2*r1*r2*np.power((r1*C1h1),gamma21)*C1h1+d2h2*r1*r2*np.power((r2*C2h2),gamma12)*C1h1+d2h2*np.power(r1,(gamma21+1))*r2*alpha21d1gamma21h1*C1h1+d2h2*np.power(r1,gamma21)*r2*alpha21d1gamma21h1*np.power((r2*C2h2),gamma12)+d2h2*np.power(r1,gamma21)*np.power(r2,(gamma12+1))*alpha21d1gamma21h1*alpha12d2gamma12h2+d2h2*np.power(r2,(gamma12+1))*alpha12d2gamma12h2*np.power((r1*C1h1),gamma21)+r1*r2*np.power((r1*C1h1),gamma21)*C1h1*C2h2+r1*r2*np.power((r2*C2h2),gamma12)*C1h1*C2h2+np.power(r1,(gamma21+1))*alpha21d1gamma21h1*np.power((r2*C2h2),gamma12)*C1h1+np.power(r2,(gamma12+1))*alpha12d2gamma12h2*np.power((r1*C1h1),gamma21)*C2h2)

        
        return U*E0 + A1*E1 + A2*E2 + (1-(U+A1+A2))*E3
    
    @staticmethod
    def _get_beta(E0, E1, E2, E3):
        """Calculates synergistic efficacy, a synergy parameter derived from E parameters.
        """
        strongest_E = np.amin(np.asarray([E1,E2]), axis=0)
        beta = (strongest_E-E3) / (E0 - strongest_E)
        return beta


    def get_parameters(self, confidence_interval=95):
        if not self._is_parameterized():
            return None
        
        #beta = (min(self.E1,self.E2)-self.E3) / (self.E0 - min(self.E1,self.E2))
        beta = MuSyC._get_beta(self.E0, self.E1, self.E2, self.E3)

        if self.converged and self.bootstrap_parameters is not None:
            parameter_ranges = self.get_parameter_range(confidence_interval=confidence_interval)
        else:
            parameter_ranges = None

        params = dict()
        params['E0'] = [self.E0, ]
        params['E1'] = [self.E1, ]
        params['E2'] = [self.E2, ]
        params['E3'] = [self.E3, ]
        params['h1'] = [self.h1, ]
        params['h2'] = [self.h2, ]
        params['C1'] = [self.C1, ]
        params['C2'] = [self.C2, ]
        params['beta'] = [beta, ]
        params['alpha12'] = [self.alpha12, ]
        params['alpha21'] = [self.alpha21, ]
        if self.variant != "no_gamma":
            params['gamma12'] = [self.gamma12, ]
            params['gamma21'] = [self.gamma21, ]

        if parameter_ranges is not None:
            params['E0'].append(parameter_ranges[:,0])
            params['E1'].append(parameter_ranges[:,1])
            params['E2'].append(parameter_ranges[:,2])
            params['E3'].append(parameter_ranges[:,3])
            params['h1'].append(parameter_ranges[:,4])
            params['h2'].append(parameter_ranges[:,5])
            params['C1'].append(parameter_ranges[:,6])
            params['C2'].append(parameter_ranges[:,7])
            params['alpha12'].append(parameter_ranges[:,8])
            params['alpha21'].append(parameter_ranges[:,9])
            if self.variant != "no_gamma":
                params['gamma12'].append(parameter_ranges[:,10])
                params['gamma21'].append(parameter_ranges[:,11])

            bsE0 = self.bootstrap_parameters[:,0]
            bsE1 = self.bootstrap_parameters[:,1]
            bsE2 = self.bootstrap_parameters[:,2]
            bsE3 = self.bootstrap_parameters[:,3]
            beta_bootstrap = MuSyC._get_beta(bsE0, bsE1, bsE2, bsE3)

            beta_bootstrap = np.percentile(beta_bootstrap, [(100-confidence_interval)/2, 50+confidence_interval/2])
            params['beta'].append(beta_bootstrap)    
        return params
    
    def summary(self, confidence_interval=95, tol=0.01):
        pars = self.get_parameters(confidence_interval=confidence_interval)
        if pars is None:
            return None
        
        ret = []
        keys = pars.keys()
        # beta
        for key in keys:
            if "beta" in key:
                l = pars[key]
                if len(l)==1:
                    if l[0] < -tol:
                        ret.append("%s\t%0.2f\t(<0) antagonistic"%(key, l[0]))
                    elif l[0] > tol:
                        ret.append("%s\t%0.2f\t(>0) synergistic"%(key, l[0]))
                else:
                    v = l[0]
                    lb,ub = l[1]
                    if v < -tol and lb < -tol and ub < -tol:
                        ret.append("%s\t%0.2f\t(%0.2f,%0.2f)\t(<0) antagonistic"%(key, v,lb,ub))
                    elif v > tol and lb > tol and ub > tol:
                        ret.append("%s\t%0.2f\t(%0.2f,%0.2f)\t(>0) synergistic"%(key, v,lb,ub))
        # alpha
        for key in keys:
            if "alpha" in key:
                l = pars[key]
                if len(l)==1:
                    if np.log10(l[0]) < -tol:
                        ret.append("%s\t%0.2f\t(<1) antagonistic"%(key, l[0]))
                    elif np.log10(l[0]) > tol:
                        ret.append("%s\t%0.2f\t(>1) synergistic"%(key, l[0]))
                else:
                    v = l[0]
                    lb,ub = l[1]
                    if np.log10(v) < -tol and np.log10(lb) < -tol and np.log10(ub) < -tol:
                        ret.append("%s\t%0.2f\t(%0.2f,%0.2f)\t(<1) antagonistic"%(key, v,lb,ub))
                    elif np.log10(v) > tol and np.log10(lb) > tol and np.log10(ub) > tol:
                        ret.append("%s\t%0.2f\t(%0.2f,%0.2f)\t(>1) synergistic"%(key, v,lb,ub))

        # gamma
        for key in keys:
            if "gamma" in key:
                l = pars[key]
                if len(l)==1:
                    if np.log10(l[0]) < -tol:
                        ret.append("%s\t%0.2f\t(<1) antagonistic"%(key, l[0]))
                    elif np.log10(l[0]) > tol:
                        ret.append("%s\t%0.2f\t(>1) synergistic"%(key, l[0]))
                else:
                    v = l[0]
                    lb,ub = l[1]
                    if np.log10(v) < -tol and np.log10(lb) < -tol and np.log10(ub) < -tol:
                        ret.append("%s\t%0.2f\t(%0.2f,%0.2f)\t(<1) antagonistic"%(key, v,lb,ub))
                    elif np.log10(v) > tol and np.log10(lb) > tol and np.log10(ub) > tol:
                        ret.append("%s\t%0.2f\t(%0.2f,%0.2f)\t(>1) synergistic"%(key, v,lb,ub))
        if len(ret)>0:
            return "\n".join(ret)
        else:
            return "No synergy or antagonism detected with %d percent confidence interval"%(int(confidence_interval))

    def __repr__(self):
        if not self._is_parameterized(): return "MuSyC()"
        
        #beta = (min(self.E1,self.E2)-self.E3) / (self.E0 - min(self.E1,self.E2))
        beta = MuSyC._get_beta(self.E0, self.E1, self.E2, self.E3)

        if self.variant == "no_gamma":
            return "MuSyC(E0=%0.2f, E1=%0.2f, E2=%0.2f, E3=%0.2f, h1=%0.2f, h2=%0.2f, C1=%0.2e, C2=%0.2e, alpha12=%0.2f, alpha21=%0.2f, beta=%0.2f)"%(self.E0, self.E1, self.E2, self.E3, self.h1, self.h2, self.C1, self.C2, self.alpha12, self.alpha21, beta)
        return "MuSyC(E0=%0.2f, E1=%0.2f, E2=%0.2f, E3=%0.2f, h1=%0.2f, h2=%0.2f, C1=%0.2e, C2=%0.2e, alpha12=%0.2f, alpha21=%0.2f, beta=%0.2f, gamma12=%0.2f, gamma21=%0.2f)"%(self.E0, self.E1, self.E2, self.E3, self.h1, self.h2, self.C1, self.C2, self.alpha12, self.alpha21, beta, self.gamma12, self.gamma21)


#    Copyright (C) 2020 David J. Wooten
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np

from ..single import Hill
from .nonparametric_base import DoseDependentHigher

class Schindler(DoseDependentHigher):
    """Schindler's multidimensional Hill equation model.
    
    From "Theory of synergistic effects: Hill-type response surfaces as 'null-interaction' models for mixtures" - Michael Schindler. This model is built to satisfy the Loewe additivity criterion, 

    Schindler assumes each drug has a Hill dose response, and defines a multidimensional Hill equation that satisfies Loewe additivity. Unlike Loewe, Schindler can be defined for combinations whose effect exceeds Emax of either drug (Loewe is limited by Emax of the *weaker* drug).

    Synergy is simply defined as the difference between the observed E and the E predicted by Schindler's multidimensional Hill equation.

    synergy : array_like, float
        (-inf,0)=antagonism, (0,inf)=synergism
    """
    
    def fit(self, d, E, single_models=None, **kwargs):
        d = np.asarray(d)
        E = np.asarray(E)
        n = d.shape[1]
        super().fit(d, E, single_models=single_models, **kwargs)
        single_models=self.single_models
        E0 = 0
        
        # Fit single drugs
        for single in single_models:
            E0 += single.E0 / n

        with np.errstate(divide='ignore', invalid='ignore'):
            # Schindler assumes drugs start at 0 and go up to Emax
            uE = E0 - E
            uE_schindler = self._model(d, E0)
            self.synergy = uE - uE_schindler

        # Ensure all single-drug Schindler scores are 0
        for i in range(n):
                # Mask where all other drugs are minimum (ideally 0)
                mask = d[:,i]>=0 # This should always be true
                for j in range(n):
                    if i==j: continue
                    mask = mask & (d[:,j]==np.min(d[:,j]))
                mask = np.where(mask)
                self.synergy[mask] = 0

        return self.synergy


    def _model(self, d, E0):
        """
        From "Theory of synergistic effects: Hill-type response surfaces as 'null-interaction' models for mixtures" - Michael Schindler
        
        E - u_hill = 0 : Additive
        E - u_hill > 0 : Synergistic
        E - u_hill < 0 : Antagonistic
        """
        h = np.asarray([model.h for model in self.single_models])
        C = np.asarray([model.C for model in self.single_models])
        Emax = E0 - np.asarray([model.Emax for model in self.single_models])

        m = d/C
        
        y = (h*m).sum(axis=1) / m.sum(axis=1)
        u_max = (Emax*m).sum(axis=1) / m.sum(axis=1)
        power = np.power(m.sum(axis=1), y)
        
        return u_max * power / (1. + power)

    def _get_single_drug_classes(self):
        return Hill, Hill


#    Copyright (C) 2020 David J. Wooten
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np

from ..single import MarginalLinear
from .nonparametric_base import DoseDependentHigher

class HSA(DoseDependentHigher):
    """Highest single agent (HSA)

    HSA says that any improvement a combination gives over the strongest single agent counts as synergy.
    """
    
    def fit(self, d, E, single_models=None, **kwargs):
        d = np.asarray(d)
        E = np.asarray(E)
        n = d.shape[1]

        super().fit(d, E, single_models=single_models, **kwargs)
        
        # Fit single drugs
        if single_models is None:
            single_models = []
            
            for i in range(n):
                # Mask where all other drugs are minimum (ideally 0)
                mask = d[:,i]>=0 # This should always be true
                for j in range(n):
                    if i==j: continue
                    mask = mask & (d[:,j]==np.min(d[:,j]))
                mask = np.where(mask)
                single = MarginalLinear()
                single.fit(d[mask,i].flatten(), E[mask], **kwargs)
                single_models.append(single)
        self.single_models = single_models

        E_singles = d*0
        for i in range(n):
            E_singles[:,i] = single_models[i].E(d[:,i])
        E_HSA = np.min(E_singles, axis=1)

        # Calculate synergy as excess over HSA
        self.synergy = E_HSA - E

        # Ensure all single-drug HSA scores are 1
        for i in range(n):
                # Mask where all other drugs are minimum (ideally 0)
                mask = d[:,i]>=0 # This should always be true
                for j in range(n):
                    if i==j: continue
                    mask = mask & (d[:,j]==np.min(d[:,j]))
                mask = np.where(mask)
                self.synergy[mask] = 0

        return self.synergy

    def _get_single_drug_classes(self):
        return MarginalLinear, None


#    Copyright (C) 2020 David J. Wooten
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


import numpy as np
from abc import ABC, abstractmethod
import inspect


from ..utils import plots
from .. import utils


class DoseDependentHigher(ABC):
    """These are models for which synergy is defined independently at each individual dose.
    """
    def __init__(self, E_bounds=(-np.inf,np.inf), h_bounds=(0,np.inf), C_bounds=(0,np.inf)):
        """Creates a DoseDependentModel

        """
        self.synergy = None
        self.d = None
        self.single_models = None

        self.E_bounds = E_bounds
        self.h_bounds = h_bounds
        self.C_bounds = C_bounds

    def fit(self, d, E, single_models=None, **kwargs):
        """Calculates dose-dependent synergy at doses d1, d2.

        Parameters
        ----------
        d : numpy.ndarray (M x N)
            Doses of N drugs sampled at M points
        
        E : array_like with length equal to M
            Dose-response at doses d

        single_models : class, array_like with length equal to N
            Model(s) to use for single drugs. If a class is given, that class is used to instantiate models for each drug. If an array is given, each element of that array must be a model that will be used for the corresponding drug.
        
        kwargs
            kwargs to pass to Hill.fit() (or whichever single-drug model is used)

        Returns
        ----------
        synergy : array_like
            The synergy calculated at all doses d
        """
        self.d = d
        self.synergy = 0*E
        self.synergy[:] = np.nan

        N = d.shape[1]

        # Initialize single drug models
        
        default_class, expected_superclass = self._get_single_drug_classes()
        
        # If an array is given, sanitize each element of the array
        if hasattr(single_models, "__iter__"):
            self.single_models = [utils.sanitize_single_drug_model(_model, default_class, expected_superclass=expected_superclass, E0_bounds=self.E_bounds, Emax_bounds=self.E_bounds, h_bounds=self.h_bounds, C_bounds=self.C_bounds) for _model in single_models]

        # Otherwise, run sanitize N times, generating N models to use
        else:
            self.single_models = []
            for _i in range(N):
                self.single_models.append(utils.sanitize_single_drug_model(single_models, default_class, expected_superclass=expected_superclass, E0_bounds=self.E_bounds, Emax_bounds=self.E_bounds, h_bounds=self.h_bounds, C_bounds=self.C_bounds))
             
        # Fit all single models if they were not given pre-fit
        for i in range(N):
            single = self.single_models[i]
            if not single.is_fit():
                # Mask where all other drugs are minimum (ideally 0)
                mask = d[:,i]>=0 # This should always be true
                for j in range(N):
                    if i==j: continue
                    mask = mask & (d[:,j]==np.min(d[:,j]))
                mask = np.where(mask)
                
                single.fit(d[mask,i].flatten(), E[mask], **kwargs)

        return self.synergy

    @abstractmethod
    def _get_single_drug_classes(self):
        """
        Returns
        -------
        default_single_class : class
            The default class type to use for single-drug models

        expected_single_superclass : class
            The required type for single-drug models. If a single-drug model is passed that is not an instance of this superclass, it will be re-instantiated using default_model
        """
        pass

    def plotly_isosurfaces(self, drug_axes=[0,1,2], other_drug_slices=None, cmap="PRGn", neglog=False, **kwargs):
        mask = self.d[:,0]>0
        n = self.d.shape[1]
        for i in range(n):
            if i in drug_axes:
                continue
            if other_drug_slices is None:
                dslice = np.min(self.d[:,i])
            else:
                dslice = other_drug_slices[i]
            mask = mask & (self.d[:,i]==dslice)

        d1 = self.d[mask,drug_axes[0]]
        d2 = self.d[mask,drug_axes[1]]
        d3 = self.d[mask,drug_axes[2]]

        Z = self.synergy[mask]
        if neglog:
            Z = -np.log10(Z)

        plots.plotly_isosurfaces(d1, d2, d3, Z, cmap=cmap, center_on_zero=True, **kwargs)


#    Copyright (C) 2020 David J. Wooten
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np

from .. import utils
from ..single import Hill_2P
from .parametric_base import ParametricModel

class Zimmer(ParametricModel):
    """The Effective Dose Model from Zimmer et al (doi: 10.1073/pnas.1606301113). This model uses the multiplicative survival principle (i.e., Bliss), but adds a parameter for each drug describing how it affects the potency of the other. Specifically, given doses d1 and d2, this model translates them to "effective" doses using the following system of equations

                            d1
    d1_eff =  --------------------------------
              1 + a12*(1/(1+(d2_eff/C2)^(-1)))

                            d2
    d2_eff =  --------------------------------
              1 + a21*(1/(1+(d1_eff/C1)^(-1)))

    Synergy by Zimmer is described by these parameters

    a12 : float
        (-(1+(d2_eff/C2))/(d2_eff/C2),0)=synergism, (0,inf)=antagonism.         Describes how drug 2 affects the effective dose of drug 1.

    a21 : float
        (-(1+(d1_eff/C1))/(d1_eff/C1),0)=synergism, (0,inf)=antagonism. Describes how drug 1 affects the effective dose of drug 2.
        
    """
    def __init__(self, h1_bounds=(0,np.inf), h2_bounds=(0,np.inf),  \
            C1_bounds=(0,np.inf), C2_bounds=(0,np.inf), a12_bounds=(-np.inf, np.inf), a21_bounds=(-np.inf, np.inf), h1=None, h2=None, C1=None, C2=None, a12=None, a21=None):

        super().__init__()
        self.C1_bounds = C1_bounds
        self.C2_bounds = C2_bounds
        self.h1_bounds = h1_bounds
        self.h2_bounds = h2_bounds
        self.a12_bounds = a12_bounds
        self.a21_bounds = a21_bounds

        with np.errstate(divide='ignore'):
            self.logh1_bounds = (np.log(h1_bounds[0]), np.log(h1_bounds[1]))
            self.logC1_bounds = (np.log(C1_bounds[0]), np.log(C1_bounds[1]))
            self.logh2_bounds = (np.log(h2_bounds[0]), np.log(h2_bounds[1]))
            self.logC2_bounds = (np.log(C2_bounds[0]), np.log(C2_bounds[1]))

        self.h1 = h1
        self.h2 = h2
        self.C1 = C1
        self.C2 = C2
        self.a12 = a12
        self.a21 = a21

        self.fit_function = lambda d, logh1, logh2, logC1, logC2, a12, a21: self._model(d[0], d[1], np.exp(logh1), np.exp(logh2), np.exp(logC1), np.exp(logC2), a12, a21)

        self.bounds = tuple(zip(self.logh1_bounds, self.logh2_bounds, self.logC1_bounds, self.logC2_bounds, self.a12_bounds, self.a21_bounds))


    def _get_initial_guess(self, d1, d2, E, drug1_model, drug2_model, p0=None):
        
        # If there is no intial guess, use single-drug models to come up with intitial guess
        if p0 is None:
            # Sanitize single-drug models
            default_class, expected_superclass = self._get_single_drug_classes()

            drug1_model = utils.sanitize_single_drug_model(drug1_model, default_class, expected_superclass=expected_superclass, h_bounds=self.h1_bounds, C_bounds=self.C1_bounds)

            drug2_model = utils.sanitize_single_drug_model(drug2_model, default_class, expected_superclass=expected_superclass, h_bounds=self.h2_bounds, C_bounds=self.C2_bounds)

            # Fit the single drug models if they were not pre-fit by the user
            if not drug1_model.is_fit():
                mask = np.where(d2==min(d2))
                drug1_model.fit(d1[mask], E[mask])
            if not drug2_model.is_fit():
                mask = np.where(d1==min(d1))
                drug2_model.fit(d2[mask], E[mask])
            
            # Get initial guesses of E0, E1, E2, h1, h2, C1, and C2 from single-drug fits
            h1, C1 = drug1_model.get_parameters()
            h2, C2 = drug2_model.get_parameters()
                        
            p0 = [h1, h2, C1, C2, 0, 0]
            
        p0 = list(self._transform_params_to_fit(p0))
        utils.sanitize_initial_guess(p0, self.bounds)
        return p0

    def _transform_params_from_fit(self, params):
        logh1, logh2, logC1, logC2, a12, a21 = params
        h1 = np.exp(logh1)
        h2 = np.exp(logh2)
        C1 = np.exp(logC1)
        C2 = np.exp(logC2)
        return h1, h2, C1, C2, a12, a21

    def _transform_params_to_fit(self, params):
        h1, h2, C1, C2, a12, a21 = params

        logh1 = np.log(h1)
        logh2 = np.log(h2)
        logC1 = np.log(C1)
        logC2 = np.log(C2)
        
        return logh1, logh2, logC1, logC2, a12, a21
        

    def _set_parameters(self, popt):
        self.h1, self.h2, self.C1, self.C2, self.a12, self.a21 = popt

    def E(self, d1, d2):
        if not self._is_parameterized():
            # ERROR
            return 0
        return self._model(d1, d2, self.h1, self.h2, self.C1, self.C2, self.a12, self.a21)

    def _reference_E(self, d1, d2):
        if not self._is_parameterized():
            return None
        return self._model(d1, d2, self.h1, self.h2, self.C1, self.C2, 0, 0)

    def _model(self, d1, d2, h1, h2, C1, C2, a12, a21):
        A = d2 + C2*(a21+1) + d2*a12
        B = d2*C1 + C1*C2 + a12*d2*C1 - d1*(d2+C2*(a21+1))
        C = -d1*(d2*C1 + C1*C2)

        d1p = (-B + np.sqrt(np.power(B,2.) - 4*A*C)) / (2.*A)
        d2p = d2 / (1. + a21 / (1. + C1/d1p))
        
        return (1 - np.power(d1p,h1)/(np.power(C1,h1)+np.power(d1p,h1))) * (1 - np.power(d2p,h2)/(np.power(C2,h2)+np.power(d2p,h2)))

    def _get_parameters(self):
        return self.h1, self.h2, self.C1, self.C2, self.a12, self.a21

    def _get_single_drug_classes(self):
        return Hill_2P, Hill_2P

    def get_parameters(self, confidence_interval=95):
        if not self._is_parameterized():
            return None
        
        if self.converged and self.bootstrap_parameters is not None:
            parameter_ranges = self.get_parameter_range(confidence_interval=confidence_interval)
        else:
            parameter_ranges = None

        params = dict()
        params['h1'] = [self.h1, ]
        params['h2'] = [self.h2, ]
        params['C1'] = [self.C1, ]
        params['C2'] = [self.C2, ]
        params['a12'] = [self.a12, ]
        params['a21'] = [self.a21, ]
        
        if parameter_ranges is not None:
            params['h1'].append(parameter_ranges[:,0])
            params['h2'].append(parameter_ranges[:,1])
            params['C1'].append(parameter_ranges[:,2])
            params['C2'].append(parameter_ranges[:,3])
            params['a12'].append(parameter_ranges[:,4])
            params['a21'].append(parameter_ranges[:,5])

        return params
    
    def summary(self, confidence_interval=95, tol=0.01):
        pars = self.get_parameters(confidence_interval=confidence_interval)
        if pars is None:
            return None
        
        ret = []
        for key in ['a12','a21']:
            l = pars[key]
            if len(l)==1:
                if l[0] < -tol:
                    ret.append("%s\t%0.2f\t(<0) synergistic"%(key, l[0]))
                elif l[0] > tol:
                    ret.append("%s\t%0.2f\t(>0) antagonistic"%(key, l[0]))
            else:
                v = l[0]
                lb,ub = l[1]
                if v < -tol and lb < -tol and ub < -tol:
                    ret.append("%s\t%0.2f\t(%0.2f,%0.2f)\t(<0) synergistic"%(key, v,lb,ub))
                elif v > tol and lb > tol and ub > tol:
                    ret.append("%s\t%0.2f\t(%0.2f,%0.2f)\t(>0) antagonistic"%(key, v,lb,ub))

        if len(ret)>0:
            return "\n".join(ret)
        else:
            return "No synergy or antagonism detected with %d percent confidence interval"%(int(confidence_interval))

    def __repr__(self):
        if not self._is_parameterized(): return "Zimmer()"
        return "Zimmer(h1=%0.2f, h2=%0.2f, C1=%0.2e, C2=%0.2e, a12=%0.2f, a21=%0.2f)"%(self.h1, self.h2, self.C1, self.C2, self.a12, self.a21)


#    Copyright (C) 2020 David J. Wooten
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import inspect
import warnings

from ..single import Hill, Hill_2P
from .nonparametric_base import DoseDependentModel
from .. import utils

class ZIP(DoseDependentModel):
    """The Zero Interaction Potency (ZIP) model (doi: 10.1016/j.csbj.2015.09.001). This model is based on the multiplicative survival principal (i.e., Bliss). All across the dose-response surface, it fits Hill-equations either holding d1==constant or d2==constant.
    
    ZIP quantifies changes in the EC50 (C) and Hill-slope (h) of each drug, as the other is increased. For instance, at a given d1==D1, d2==D2, ZIP fits two Hill equations: one for drug1 (fixing d2==D2), and one for drug2 (fixing d1==D1). These Hill equations are averaged to get a fit value of E (y_c in their paper) at these doses. This is then subtracted from the expected result (y_zip in their paper) obtained by assuming these Hill equations have equivalent h and C to their single-drug counterparts. This difference (delta in their paper) becomes the metric of synergy.

    ZIP models store these delta values as model._synergy, but also store the Hill equation fits for drug1 and drug2 across the whole surface, allowing investigation of how h and C change across the surface

    synergy : array_like
        (-inf,0)=antagonism, (0,inf)=synergism. The "delta" synergy score from ZIP

    _h_21 : array_like
        The hill slope of drug 1 obtained by holding D2==constant

    _h_12 : array_like
        The hill slope of drug 2 obtained by holding D1==constant

    _C_21 : array_like
        The EC50 of drug 1 obtained by holding D2==constant

    _C_12 : array_like
        The EC50 of drug 2 obtained by holding D1==constant
    """
    def __init__(self, E0_bounds=(0,1.5), E1_bounds=(0,1.5), E2_bounds=(0,1.5), h1_bounds=(0,np.inf), C1_bounds=(0,np.inf), h2_bounds=(0,np.inf), C2_bounds=(0,np.inf), synergyfinder=False):

        super().__init__(h1_bounds=h1_bounds, h2_bounds=h2_bounds, C1_bounds=C1_bounds, C2_bounds=C2_bounds, E0_bounds=E0_bounds, E1_bounds=E1_bounds, E2_bounds=E2_bounds)

        self.synergyfinder = synergyfinder

        self._h_21 = []
        self._h_12 = []
        self._C_21 = []
        self._C_12 = []
        self._Emax_21 = []
        self._Emax_12 = []
        
    def _get_single_drug_classes(self):
        return Hill, Hill

    def fit(self, d1, d2, E, drug1_model=None, drug2_model=None, use_jacobian=True, **kwargs):
    
        d1 = np.asarray(d1)
        d2 = np.asarray(d2)
        E = np.asarray(E)
        super().fit(d1,d2,E, drug1_model=drug1_model, drug2_model=drug2_model, use_jacobian=use_jacobian, **kwargs)

        drug1_model = self.drug1_model
        drug2_model = self.drug2_model
 
        E0_1, Emax_1, h1, C1 = drug1_model.get_parameters()
        E0_2, Emax_2, h2, C2 = drug2_model.get_parameters()
        E0 = (E0_1+E0_2)/2.
        Emax = (Emax_1+Emax_2)/2.

        if (E0 < Emax):
            drug1_model.E0 = Emax_1
            drug2_model.E0 = Emax_2
            drug1_model.Emax = E0_1
            drug2_model.Emax = E0_2

            tmp = E0
            E0 = Emax
            Emax = tmp
        
        logh1 = np.log(h1)
        logh2 = np.log(h2)
        logC1 = np.log(C1)
        logC2 = np.log(C2)

        self._h_21 = [] # Hill slope of drug 1, after treated by drug 2
        self._h_12 = [] # Hill slope of drug 2, after treated by drug 1
        self._C_21 = [] # EC50 of drug 1, after treated by drug 2
        self._C_12 = [] # EC50 of drug 2, after treated by drug 1
        self._Emax_21 = []
        self._Emax_12 = []
        
        if self.synergyfinder:
            zip_model = _Hill_3P(Emax_bounds=(-1e-6,1e-6))
        else:
            zip_model = _Hill_3P(Emax_bounds=(0,1.5))
        

        for D1, D2 in zip(d1, d2):
            # Fix d2==D2, and fit hill for D1
            mask = np.where(d2==D2)
            y2 = drug2_model.E(D2)
            zip_model.E0 = y2
            zip_model.fit(d1[mask],E[mask], use_jacobian=use_jacobian, p0=[Emax_1,h1,C1])
            self._h_21.append(zip_model.h)
            self._C_21.append(zip_model.C)
            self._Emax_21.append(zip_model.Emax)

            # Fix d1==D1, and fit hill for D2
            mask = np.where(d1==D1)
            y1 = drug1_model.E(D1)
            zip_model.E0 = y1
            zip_model.fit(d2[mask],E[mask], use_jacobian=use_jacobian, p0=[Emax_2,h2,C2])
            self._h_12.append(zip_model.h)
            self._C_12.append(zip_model.C)
            self._Emax_12.append(zip_model.Emax)
        
        self._h_21 = np.asarray(self._h_21)
        self._h_12 = np.asarray(self._h_12)
        self._C_21 = np.asarray(self._C_21)
        self._C_12 = np.asarray(self._C_12)
        self._Emax_21 = np.asarray(self._Emax_21)
        self._Emax_12 = np.asarray(self._Emax_12)

        self.synergy = self._delta_score(d1, d2, E0, self.drug1_model.Emax, self.drug2_model.Emax, h1, h2, C1, C2, self._Emax_21, self._Emax_12, self._h_21, self._h_12, self._C_21, self._C_12)

        mask = np.where((d1==0) | (d2==0))
        self.synergy[mask]=0
        self.synergy

        return self.synergy

    def _delta_score(self, d1, d2, E0, E1, E2, h1, h2, C1, C2, Emax_21, Emax_12, h_21, h_12, C_21, C_12):

        single_drug_1 = E0 + (E1-E0) * np.power(d1,h1) / (np.power(C1,h1) + np.power(d1,h1))

        single_drug_2 = E0 + (E2-E0) * np.power(d2,h2) / (np.power(C2,h2) + np.power(d2,h2))

        zip_fit_1 = single_drug_2 + (Emax_21-single_drug_2) * np.power(d1,h_21) / (np.power(C_21,h_21) + np.power(d1,h_21))

        zip_fit_2 = single_drug_1 + (Emax_12-single_drug_1) * np.power(d2,h_12) / (np.power(C_12,h_12) + np.power(d2,h_12))
        
        zip_fit = (zip_fit_1+zip_fit_2)/2.
        zip_ind = single_drug_1*single_drug_2

        self.reference = zip_ind

        return zip_ind-zip_fit



class _Hill_3P(Hill):
    def __init__(self, E0=1, Emax=0, h=None, C=None, Emax_bounds=(-np.inf, np.inf), h_bounds=(0,np.inf), C_bounds=(0,np.inf)):
        super().__init__(h=h, C=C, E0=E0, Emax=Emax, Emax_bounds=Emax_bounds, h_bounds=h_bounds, C_bounds=C_bounds)

        self.fit_function = lambda d, Emax, logh, logC: self._model(d, self.E0, Emax, np.exp(logh), np.exp(logC))

        self.jacobian_function = lambda d, Emax, logh, logC: self._model_jacobian(d, Emax, logh, logC)

        self.bounds = tuple(zip(self.Emax_bounds, self.logh_bounds, self.logC_bounds))

    def _model_jacobian(self, d, Emax, logh, logC):
        dh = d**(np.exp(logh))
        Ch = (np.exp(logC))**(np.exp(logh))
        logd = np.log(d)
        E0 = self.E0

        jEmax = dh/(Ch+dh)

        jC = (E0-Emax)*dh*np.exp(logh+logC)*(np.exp(logC))**(np.exp(logh)-1) / ((Ch+dh)*(Ch+dh))

        jh = (Emax-E0)*dh*np.exp(logh) * ((Ch+dh)*logd - (logC*Ch + logd*dh)) / ((Ch+dh)*(Ch+dh))
        
        jac = np.hstack((jEmax.reshape(-1,1), jh.reshape(-1,1), jC.reshape(-1,1)))
        jac[np.isnan(jac)]=0
        return jac

    def _get_initial_guess(self, d, E, p0=None):

        if p0 is None:
            p0 = [np.nanmin(E), 1, np.median(d)]
            
        p0 = list(self._transform_params_to_fit(p0))
        utils.sanitize_initial_guess(p0, self.bounds)
        
        return p0

    def get_parameters(self):
        """Gets the model's parameters
        
        Returns
        ----------
        parameters : tuple
            (Emax, h, C)
        """
        return (self.Emax, self.h, self.C)

    def _set_parameters(self, popt):
        Emax, h, C = popt
        
        self.Emax = Emax
        self.h = h
        self.C = C

    def _transform_params_from_fit(self, params):
        return params[0], np.exp(params[1]), np.exp(params[2])

    def _transform_params_to_fit(self, params):
        return params[0], np.log(params[1]), np.log(params[2])

    def __repr__(self):
        if not self._is_parameterized(): return "Hill_3P()"
        
        return "Hill_3P(E0=%0.2f, Emax=%0.2f, h=%0.2f, C=%0.2e)"%(self.E0, self.Emax, self.h, self.C)


#    Copyright (C) 2020 David J. Wooten
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np

from ..single import Hill_CI
from .nonparametric_base import DoseDependentHigher

class CombinationIndex(DoseDependentHigher):
    """Combination Index (CI).
    
    CI is a mass-action-derived model of drug combination synergy. In the original 1984 paper (XXX), CI was derived with two forms: (1) mutually-exclusive drugs, and (2) mutually non-exclustive drugs. Since then, model (1) has been preferred, and is the one implemented here.

    CI fits single-drug responses using a log-linearization scatchard-like regression, that implicitly assumes E0=1 and Emax=0, so should only be applied in those cases. If this assumption is not met, users may wish to use Loewe(), which is here calculated identically to CI, but without the E0 and Emax assumptions, and fits drugs to a 4-parameter Hill equation using nonlinear optimization.

    synergy : array_like, float
        [0,1)=synergism, (1,inf)=antagonism
    """

    def fit(self, d, E, single_models=None, **kwargs):
        
        d = np.asarray(d)
        E = np.asarray(E)
        n = d.shape[1]

        super().fit(d, E, single_models=single_models, **kwargs)
        
        # Fit single drugs
        if single_models is None:
            single_models = []
            
            for i in range(n):
                # Mask where all other drugs are minimum (ideally 0)
                mask = d[:,i]>=0 # This should always be true
                for j in range(n):
                    if i==j: continue
                    mask = mask & (d[:,j]==np.min(d[:,j]))
                mask = np.where(mask)
                single = Hill_CI.create_fit(d[mask,i].flatten(), E[mask])
                single_models.append(single)
        self.single_models = single_models

        d_singles = d*0
        with np.errstate(divide='ignore', invalid='ignore'):
            for i in range(n):
                d_singles[:,i] = single_models[i].E_inv(E)

            self.synergy = (d/d_singles).sum(axis=1)

        # Ensure all single-drug CI scores are 1
        for i in range(n):
                # Mask where all other drugs are minimum (ideally 0)
                mask = d[:,i]>=0 # This should always be true
                for j in range(n):
                    if i==j: continue
                    mask = mask & (d[:,j]==np.min(d[:,j]))
                mask = np.where(mask)
                self.synergy[mask] = 1

        return self.synergy

    def _get_single_drug_classes(self):
        return Hill_CI, Hill_CI

    def plotly_isosurfaces(self, drug_axes=[0,1,2], other_drug_slices=None, cmap="PRGn", neglog=True, **kwargs):
        super().plotly_isosurfaces(drug_axes=drug_axes, other_drug_slices=other_drug_slices, cmap=cmap, neglog=neglog, **kwargs)